copy %HOMEPATH%\onvif-gui-win-libs\ffmpeg\bin\* %HOMEPATH%\libonvif\libavio\avio
copy %HOMEPATH%\onvif-gui-win-libs\sdl\bin\* %HOMEPATH%\libonvif\libavio\avio
copy %HOMEPATH%\onvif-gui-win-libs\libxml2\bin\* %HOMEPATH%\libonvif\libonvif\libonvif